# Generated by Django 4.1.3 on 2022-11-09 06:43

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TaskkitControlEvent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sent', models.FloatField(db_index=True)),
                ('data', models.BinaryField()),
            ],
            options={
                'db_table': 'taskkit_control_event',
            },
        ),
        migrations.CreateModel(
            name='TaskkitLock',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
            ],
            options={
                'db_table': 'taskkit_lock',
            },
        ),
        migrations.CreateModel(
            name='TaskkitSchedulerState',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('data', models.BinaryField()),
            ],
            options={
                'db_table': 'taskkit_scheduler_state',
            },
        ),
        migrations.CreateModel(
            name='TaskkitWorker',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('expires', models.FloatField(db_index=True)),
            ],
            options={
                'db_table': 'taskkit_worker',
            },
        ),
        migrations.CreateModel(
            name='TaskkitTask',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('group', models.CharField(max_length=40)),
                ('name', models.CharField(max_length=255)),
                ('data', models.BinaryField()),
                ('due', models.FloatField()),
                ('created', models.FloatField()),
                ('scheduled', models.FloatField(null=True)),
                ('retry_count', models.PositiveIntegerField()),
                ('ttl', models.FloatField()),
                ('assignee_worker_id', models.CharField(max_length=255, null=True)),
                ('began', models.FloatField(null=True)),
                ('result', models.BinaryField(null=True)),
                ('error_message', models.TextField(null=True)),
                ('done', models.FloatField(null=True)),
                ('disposable', models.FloatField(db_index=True, null=True)),
            ],
            options={
                'db_table': 'taskkit_task',
                'index_together': {('done', 'began'), ('name', 'created'), ('began', 'group', 'due')},
            },
        ),
    ]
